#ifndef Encoding_hpp
#define Encoding_hpp

namespace SoftWire
{
	typedef unsigned char byte;

	class Synthesizer;

	class Encoding
	{
		friend Synthesizer;

	public:
		enum Reg
		{
			REG_UNKNOWN = -1,

			AL = 0, AX = 0, EAX = 0, ST0 = 0, MM0 = 0, XMM0 = 0,
			CL = 1, CX = 1, ECX = 1, ST1 = 1, MM1 = 1, XMM1 = 1,
			DL = 2, DX = 2, EDX = 2, ST2 = 2, MM2 = 2, XMM2 = 2,
			BL = 3, BX = 3, EBX = 3, ST3 = 3, MM3 = 3, XMM3 = 3,
			AH = 4, SP = 4, ESP = 4, ST4 = 4, MM4 = 4, XMM4 = 4,
			CH = 5, BP = 5, EBP = 5, ST5 = 5, MM5 = 5, XMM5 = 5,
			DH = 6, SI = 6, ESI = 6, ST6 = 6, MM6 = 6, XMM6 = 6,
			BH = 7, DI = 7, EDI = 7, ST7 = 7, MM7 = 7, XMM7 = 7
		};
		
		Encoding();

		~Encoding();

		void reset();

		const char *getLabel() const;
		const char *getReference() const;

		void addPrefix(byte p);
		void checkFormat() const;

		int length() const;   // Length of encoded instruction in bytes
		int writeCode(byte *output) const;

		void setImmediate(int immediate);
		void setJumpOffset(int offset);
		void setLabel(const char *label);
		void setReference(const char *label);

	private:
		char *label;
		char *reference;

		enum Mod
		{
			MOD_NO_DISP = 0,
			MOD_BYTE_DISP = 1,
			MOD_DWORD_DISP = 2,
			MOD_REG = 3
		};

		enum Scale
		{
			SCALE_UNKNOWN = 0,
			SCALE_1 = 0,
			SCALE_2 = 1,
			SCALE_4 = 2,
			SCALE_8 = 3
		};

		struct
		{
			bool P1 : 1;
			bool P2 : 1;
			bool P3 : 1;
			bool P4 : 1;
			bool O2 : 1;
			bool O1 : 1;
			bool modRM : 1;
			bool SIB : 1;
			bool D1 : 1;
			bool D2 : 1;
			bool D3 : 1;
			bool D4 : 1;
			bool I1 : 1;
			bool I2 : 1;
			bool I3 : 1;
			bool I4 : 1;
			
		} format;

		byte P1;   // Prefixes
		byte P2;
		byte P3;
		byte P4;
		byte O1;   // Opcode
		byte O2;
		struct
		{
			union
			{
				struct
				{
					byte r_m : 3;
					byte reg : 3;
					byte mod : 2;
				};

				byte b;
			};
		} modRM;
		struct
		{
			union
			{
				struct
				{
					byte base : 3;
					byte index : 3;
					byte scale : 2;
				};

				byte b;
			};
		} SIB;
		union
		{
			int displacement;

			struct
			{
				byte D1;
				byte D2;
				byte D3;
				byte D4;
			};
		};
		union
		{
			int immediate;

			struct
			{
				byte I1;
				byte I2;
				byte I3;
				byte I4;
			};
		};
	};
}

#endif   // Encoding_hpp